#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# XOR, 排他的論理和を学習する２層ニューラルネットワーク
# 活性化関数をReLUに差し替えてみる

import numpy as np

# 1個のニューロンを表わすクラス
class Neuron:
    # input_size : ニューロンの入力数
    # kind : ニューロンの種類を指定する、0=線形ニューロン, 1=シグモイドニューロン, それ以外=ReLU
    def __init__(self, input_size=2, kind=0):
        # 重み W を -1 ～ 1 の乱数で初期化する
        self.W = np.random.rand(input_size) * 2 - 1
        # 閾値 B を -1 ～ 1 の乱数で初期化する
        self.B = np.random.rand() * 2 - 1
        if kind == 0:   # kindに応じて活性化関数をセットする
            self.activate, self.d_active = self.ident, self.d_ident # 恒等関数をセット
        elif kind==1:
            self.activate, self.d_active = self.sigm, self.d_sigm   # シグモイド関数をセット
        else:
            self.activate, self.d_active = self.relu, self.d_relu   # ReLU関数をセット
    
    def ident(self, x): # 恒等関数
        return x    # 入力値をそのまま返す
    
    def d_ident(self, y):  # 恒等関数の微分
        return np.ones(y.shape)  # yと同じ形で 1を返す
    
    def sigm(self, x):  # シグモイド関数
        return  1 / (1 + np.exp( -x ))
    
    def d_sigm(self, y): # シグモイド関数の微分
        return y * (1 - y)
    
    def relu(self, x):   # ReLU関数
        return np.where( x <= 0, 0, x )
    
    def d_relu(self, y): # ReLU関数の微分
        return np.where( y <= 0, 0, 1 )
    
    def forward(self, x):   # 順方向の計算
        u = np.sum(x * self.W) + self.B
            # 複数のデータxと重みWの掛け算を一気に行なう
        y = self.activate(u)    # 活性化関数を施す
        return y

LEARNING_RATE = 0.01    # 学習係数 -- ここをうまく調整する
MIDDLE_SIZE = 4         # 中間層のニューロンの数
    # シグモイドの場合 2 で十分だが、ReLUの場合 4 にする必要がある

mid_layer = [ Neuron(input_size=2, kind=2) for _ in range(MIDDLE_SIZE) ]
    # ニューロンをMIDDLE_SIZE個並べた中間層
    # kind=1でシグモイド、kind=2でReLU
out_layer = [ Neuron(input_size=MIDDLE_SIZE, kind=0) ] 
    # 線形ニューロン1個の出力層

mid_outputs = np.zeros(MIDDLE_SIZE)   # 中間層の出力を保持しておく
tmp_weights = np.zeros(MIDDLE_SIZE)   # 一時的に出力層ニューロンの重みを保持しておく

for trial in range( 200000 ):   # 繰り返しトレーニングする
    # 訓練データを用意
    x = np.random.randint(0, 1+1, 2)    # 0,1の乱数を2個生成する
    y = x[0] ^ x[1]     # XORをとって教師データをつくる
    
    # 順方向伝播 -- 中間層
    for i, mid_neuro in enumerate( mid_layer ):
        mid_outputs[i] = mid_neuro.forward( x )
    
    # 順方向伝播 -- 出力層
    out_neuro = out_layer[0]
    result = out_neuro.forward( mid_outputs )   
    
    # ニューラルネットの出力結果
    error = result - y   # 誤差 = 教師データと結果の差分
    
    if trial % 100 == 0: # 100回に1回、結果を表示する
        print( "x={}, y={}, result={}, error={}".format( x, y, result, error ) )
    
    # 逆方向伝播 -- 出力層
    tmp_weights = out_neuro.W
        # 次の中間層の計算のため、更新前の値を保持しておく
    out_neuro.W -= ( LEARNING_RATE * error * mid_outputs )
        # 出力層ニューロンの重みを更新
    out_neuro.B -= ( LEARNING_RATE * error )
        # 出力層ニューロンの閾値を更新
    
    # 逆方向伝播 -- 中間層
    for i, mid_neuro in enumerate( mid_layer ):
        back_delta = mid_neuro.d_active( mid_outputs[i] ) * error * tmp_weights[i]
        mid_neuro.W -= ( LEARNING_RATE * back_delta * x )   # 中間層ニューロンの重みを更新
        mid_neuro.B -= ( LEARNING_RATE * back_delta )       # 中間層ニューロンの閾値を更新

