#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MNIST

import numpy as np

# 1枚の層を表わすクラス
class Layer:
    
    LEARNING_RATE = 0.05    # 学習係数 -- ここをうまく調整する
    
    def __init__(self, input_size, layer_size, kind=0):
        self.input_size = input_size    # 入力の数
        self.layer_size = layer_size    # ニューロンの数
        
        # kindに応じて活性化関数とその微分をセットする
        if kind == 0:   # 線形ニューロンとする
            self.activate, self.d_active = self._ident, self._d_ident
            sd = np.sqrt(2/(self.input_size + self.layer_size)) # Xavierの初期化、標準偏差
        elif kind == 1: # シグモイドニューロンとする
            self.activate, self.d_active = self._sigm, self._d_sigm
            sd = np.sqrt(2/(self.input_size + self.layer_size)) # Xavierの初期化、標準偏差
        elif kind == 2: # ReLUを設定
            self.activate, self.d_active = self._relu, self._d_relu
            sd = np.sqrt(2/self.input_size) # Heの初期化、標準偏差
        else:   # 出力用にソフトマックス関数を設定
            self.activate, self.d_active = self._softmax, self._d_softmax
            sd = np.sqrt(2/(self.input_size + self.layer_size)) # Xavierの初期化、標準偏差
        
        self.W = np.random.normal( 0, sd, (self.layer_size, self.input_size) )  # 重み、ニューロン数×入力数
        self.B = np.zeros( self.layer_size )    # 閾値、ニューロン数だけある。元文献だとバイアスは0としている
    
    def _ident(self, x):    # 恒等関数
        return x    # 入力値をそのまま返す
    
    def _d_ident(self, y):  # 恒等関数の微分
        return np.ones(y.shape) # yと同じ形式で 1を返す
    
    def _sigm(self, x): # シグモイド関数
        return  1 / ( 1 + np.exp( -x ) )
    
    def _d_sigm(self, y):   # シグモイド関数の微分
        return y * (1.0 - y)
    
    def _relu(self, x): # ReLU
        return np.where( x <= 0, 0, x )
    
    def _d_relu(self, y):   # ReLUの微分
        return np.where( y <= 0, 0, 1 )
    
    def _softmax(self, x):  # ソフトマックス関数
        if x.ndim == 2: # ミニバッチ、入力 x が行列の場合
            x = x.T # いったん縦横を転置する（集計の都合で）
            mx = np.max(x, axis=0)  # オーバーフロー対策、各試行ごとに最大値をとる
            ex = np.exp(x - mx)
            result = ex / np.sum(ex, axis=0)
            return result.T # 転置を元に戻す
        else:   # オンライン、入力 x が1件だけの場合
            mx = np.max(x)  # オーバーフロー対策
            ex = np.exp(x - mx)
            return ex / np.sum(ex)
    
    def _d_softmax(self, y):
        return np.ones(y.shape)  # yと同じ形式で 1を返す（ソフトマックス+交差エントロピー誤差だとこうなる）
    
    def forward(self, x):   # 順方向伝播の計算
        self.input = x  # 入力値を保持しておく
        u = x @ self.W.T + self.B  # @ は行列の掛け算、複数のデータxと重みWの掛け算を一気に行なう
            # このとき B はすべての行に対して足される
            #（「５－１」で「劇的に改善される」と指摘したのが、このプログラム行）
        self.output = self.activate(u)  # 活性化関数を施す
    
    def backward(self, error):  # 逆方向伝播の計算
        batch_size = error.shape[0] # 誤差がバッチサイズだけやってくる
        
        u = self.d_active(self.output)  # 活性化関数の微分
        delta = u * error   # 誤差を活性化関数以前の姿に戻す
        
        self.d_W = delta.T @ self.input  / batch_size  
            # 重みの更新値、行列の掛け算で一気に計算
            #（「５－１」で「劇的に改善される」と指摘したプログラム行）
        self.d_B = np.sum(delta, axis=0) / batch_size  # バイアスの更新値
        
        self.back_error = delta @ self.W    # 1つ手前の層に戻すべき誤差
    
    def update(self):   # 重みとバイアスの更新
        self.W -= self.LEARNING_RATE * self.d_W # 重みの更新
        self.B -= self.LEARNING_RATE * self.d_B # バイアスの更新

INPUT_SIZE = 784    # 入力の数 = 28*28 = x_train.shape[1] * x_train.shape[2]
MIDDLE_SIZE = 50    # 中間層の数 -- うまく調整する
N_CLASS = 10        # 出力の数 -- 10カテゴリーに分類する

# MNIST手書き文字データを kerasライブラリーから読み込む
from keras.datasets import mnist    # kerasデータセットを利用
(x_train, y_train), (x_test, y_test) = mnist.load_data()
n_train = x_train.shape[0]  # トレーニングデータの件数、60000件
n_test  = x_test.shape[0]   # テストデータの件数、10000件

x_train = x_train.reshape(n_train, INPUT_SIZE).astype('float32') / 255
    # INPUT_SIZEの数値ベクトルに整形
x_test  = x_test.reshape( n_test,  INPUT_SIZE).astype('float32') / 255
    # 255階調のデータを0～1に正規化

# 教師データをOne-hot化する -- 単位行列をつくって行数indexを指定するという技
y_train = np.eye(N_CLASS)[y_train]
y_test  = np.eye(N_CLASS)[y_test]

mid_layer = Layer(input_size=INPUT_SIZE,  layer_size=MIDDLE_SIZE, kind=2)   # 中間層、kind=1:シグモイド, 2:ReLU
out_layer = Layer(input_size=MIDDLE_SIZE, layer_size=N_CLASS,     kind=3)   # 出力層、kind=0:線形, 1:シグモイド, 3:softmax

BATCH_SIZE = 100  # バッチサイズ、1回のバッチで処理するデータ件数

for epoch in range( 50000 ):    # 繰り返しトレーニングする
    # 訓練データを用意 -- バッチサイズだけランダムに選び出す
    rnd = np.random.choice(n_train, BATCH_SIZE)
    x = x_train[rnd]    # 本来であれば、1度用いたデータは次回には使わない方が良い。
    y = y_train[rnd]    # ここでは簡易的な方法をとった。
    
    # 順方向伝播
    mid_layer.forward(x)    # 中間層
    out_layer.forward( mid_layer.output )   # 出力層
    
    error = out_layer.output - y    # 誤差 = 教師データと結果の差分
    
    # 訓練の経過を表示する（実はこの表示のための誤差を計算しなくても学習はできる）
    mse = np.mean( np.sum(error ** 2, axis=1) ) # ２乗誤差を求めておこう
    cx_entropy = np.mean( np.sum( - y * np.log(out_layer.output + 1e-7), axis=1) )
    print("{}: MSE={}, X_ETP={}".format( epoch, mse, cx_entropy )) if epoch % 100 == 0 else None # 100回に 1回表示
    
    # 逆方向伝播
    out_layer.backward(error) # 出力層
    out_layer.update()
    mid_layer.backward( out_layer.back_error ) # 中間層
    mid_layer.update()

# 学習結果をテストする
hit = 0
for trial, (x, y) in enumerate( zip(x_test, y_test) ):
    # 順方向伝播
    mid_layer.forward(x)    # 中間層
    out_layer.forward( mid_layer.output )   # 出力層
    
    result = np.argmax(out_layer.output)
    teach  = np.argmax(y)
    hit += (result == teach)    # 結果が当たった件数

print( "{} hits, accracy={}%".format( hit, 100 * hit / n_test ) ) # 97.4%程度の精度が実現できる
